/**
 * @file gui_w_button.c
 * @brief basic button implementation
 * @author Duboisset Philippe
 * @version 0.1b
 * @date (yyyy-mm-dd) 2013-07-13
 *
 * Copyright (C) <2013>  Duboisset Philippe <duboisset.philippe@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "gui.h"


/**
 * button
 */
typedef struct {

  uint8_t /*@null@*/ *str;  /*string associated to the button*/
  gui_img_t img;            /*image of the button*/

  /*graphic context; each button owns a dedicated colorText & font*/
  color_t colorText;
  gui_font_t font;
} button_st;


/**
 * local functions
 */
static void ButtonDraw(void *g_obj, void *obj);


/**
 * @function GUI_AddButton
 * @brief add a basic button
 * @param const rect_st *rec: object dimension
 * @param const void *str: text of the button (will be copied into it)
 * @param gui_img_t img: image of the button
 * @return g_obj_st *: pointer to the associated generic object if succedeed, NULL if error.
 */
g_obj_st /*@null@*/ *GUI_AddButton(const rect_st *rec, const void /*@null@*/ *str, gui_img_t img) {

  g_obj_st *g_obj = NULL, *res = NULL;
  button_st *button = NULL;

  /*check parameters*/
  if(rec != NULL) {

    /*allocate a generic object*/
    g_obj = GUI_AddGenericObject();
    if(g_obj != NULL) {

      /*allocate a button*/
      button = salloc(sizeof(button_st));
      if(button != NULL) {

        /*button settings*/
        button->img = img;
        button->colorText = GetColor(G_COL_TEXT);
        button->font = GetCurrentFont();
        button->str = AllocateAndCopyStr(str);

        /*linkage between generic obj & button*/
        g_obj->rec = *rec;
        g_obj->draw = ButtonDraw;
        g_obj->task = NULL;
        g_obj->obj = button;
        GUI_ObjSetFocusable(g_obj, true); /*button is a focusable object*/
        res = g_obj;
      }
    }
  }

  return res;
}


/**
 * @function ButtonDraw
 * @brief button draw function
 * @param void *_g_obj: generic object
 * @param void *_obj: button object
 * @return none
 */
static void ButtonDraw(void *_g_obj, void *_obj) {

  g_obj_st *g_obj;
  button_st *button;
  color_t colBack, colUpperRec, colLowerRec, colText;
  coord_t x0, y0, x1, y1, ox, oy, x, y;
  rect_st frec;

  /*retreive generic & specific object*/
  if(_g_obj != NULL && _obj != NULL) {
    g_obj = (g_obj_st *) _g_obj;
    button =  (button_st*) _obj;

    /*disabled object?*/
    if(GUI_ObjIsDisabled(g_obj)) {
      SetLut(G_LUT_DISABLED);
      colBack     = GetColor(G_COL_BACKGROUND);
      colUpperRec = GetColor(G_COL_UPPER_REC);
      colLowerRec = GetColor(G_COL_LOWER_REC);
      colText     = GetColor(G_COL_D_TEXT);;
      ox = oy = 0;
    }
    /*pressed object?*/
    else if(GUI_ObjIsPressed(g_obj)) {
      SetLut(G_LUT_NORMAL);
      colBack     = GetColor(G_COL_BACKGROUND);
      colUpperRec = GetColor(G_COL_LOWER_REC);
      colLowerRec = GetColor(G_COL_UPPER_REC);
      colText     = button->colorText;
      ox = oy = 1;
    }
    /*notified object?*/
    else if(GUI_ObjIsBlink(g_obj)) {
      SetLut(G_LUT_NORMAL);
      colBack     = GetColor(G_COL_SPECIAL);
      colUpperRec = GetColor(G_COL_UPPER_REC);
      colLowerRec = GetColor(G_COL_LOWER_REC);
      colText     = button->colorText;
      ox = oy = 0;
    }
    /*normal*/
    else {
      SetLut(G_LUT_NORMAL);
      colBack     = GetColor(G_COL_BACKGROUND);
      colUpperRec = GetColor(G_COL_UPPER_REC);
      colLowerRec = GetColor(G_COL_LOWER_REC);
      colText     = button->colorText;
      ox = oy = 0;
    }

    /*P2D configuration*/
    P2D_SetDisplayMode(DISPLAY_SOLID);
    P2D_SetLineType(LINE_SOLID);
    P2D_SetColors(colBack, colBack);

    /*background*/
    P2D_FillRect(&(g_obj->rec));

    (void) P2D_RectToCoord(&(g_obj->rec), &x0, &y0, &x1, &y1);

    /*text*/
    if(button->str) {
      SetFont(button->font);
      P2D_SetColor(colText);
      x = x0 + ox + 2;
      y = (y1 - y0 - P2D_GetTextHeight()) / 2 + y0 + oy;
      P2D_PutText(x, y, button->str);
    }

    /*sprite*/
    x = g_obj->rec.w - SpriteGetWidth(button->img) - 2 + x0 + ox;
    y = (g_obj->rec.h - SpriteGetHeight(button->img)) / 2 + y0 + oy;
    Sprite(x, y, button->img);

    /*upper rect*/
    P2D_SetColor(colUpperRec);
    P2D_Line(x0, y0, x1, y0);
    P2D_Line(x0, y0, x0, y1);

    /*lower rect*/
    P2D_SetColor(colLowerRec);
    P2D_Line(x0, y1, x1, y1);
    P2D_Line(x1, y0, x1, y1);

    /*focus*/
    if(GUI_ObjIsFocused(g_obj)) {
      x0 += 2;
      y0 += 2;
      x1 -= 2;
      y1 -= 2;
      P2D_SetColor(button->colorText);
      (void) P2D_CoordToRect(&frec, x0, y0, x1, y1);
      P2D_SetLineType(LINE_DOT);
      P2D_Rect(&frec);
    }
  }
}
