/**
 * @file gui_w_text.c
 * @brief text area implementation
 * @author Duboisset Philippe
 * @version 0.1b
 * @date (yyyy-mm-dd) 2013-07-13
 *
 * Copyright (C) <2013>  Duboisset Philippe <duboisset.philippe@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "gui.h"


/**
 * text area
 */
typedef struct {

  uint8_t /*@null@*/ *str;           /*string associated to the text area*/

  /*graphic context; each frame owns a dedicated colorText & font*/
  color_t colorText;
  gui_font_t font;
} text_st;


/**
 * local functions
 */
static void TextDraw(void *g_obj, void *obj);



/**
 * @function GUI_AddText
 * @brief add a text area
 * @param const rect_st *rec: object dimension
 * @param const void *str: text of the text area (will be copied into it)
 * @return g_obj_st *: pointer to the associated generic object if succedeed, NULL if error.
 */
g_obj_st /*@null@*/ *GUI_AddText(const rect_st *rec, const void /*@null@*/ *str) {

  g_obj_st *g_obj = NULL, *res = NULL;
  text_st *text = NULL;

  /*check parameters*/
  if(rec != NULL && str != NULL) {

    /*allocate a generic object*/
    g_obj = GUI_AddGenericObject();
    if(g_obj != NULL) {

      /*allocate & init the text area*/
      text = salloc(sizeof(text_st));
      if(text != NULL) {

        text->colorText = GetColor(G_COL_TEXT);
        text->font = GetCurrentFont();
        text->str = AllocateAndCopyStr(str);

        /*linkage between generic obj & text*/
        g_obj->rec = *rec;
        g_obj->draw = TextDraw;
        g_obj->task = NULL;
        g_obj->obj = text;
        GUI_ObjSetStatic(g_obj, true); /*text area is static: no need to refresh when pressed*/
        res = g_obj;
      }
    }
  }
  return res;
}


//////!!!! TODO: word wrap, auto line return, ...

/**
 * @function TextDraw
 * @brief text area draw function
 * @param void *_g_obj: generic object
 * @param void *_obj: text object
 * @return none
 */
static void TextDraw(void *_g_obj, void *_obj) {

  g_obj_st *g_obj;
  text_st *text;
  color_t colText;

  /*retreive generic & specific object*/
  if(_g_obj != NULL && _obj != NULL) {
    g_obj = (g_obj_st *) _g_obj;
    text =  (text_st*) _obj;
	
    /*disabled object?*/
    if(GUI_ObjIsDisabled(g_obj)) {
      colText = GetColor(G_COL_D_TEXT);
    }
    /*normal*/
    else {
      colText = text->colorText;
    }

    P2D_SetColors(colText, GetColor(G_COL_BACKGROUND));
    SetFont(text->font);

    /*P2D configuration*/
    P2D_SetDisplayMode(DISPLAY_SOLID);
    P2D_SetLineType(LINE_SOLID);

    P2D_PutText(g_obj->rec.x, g_obj->rec.y, text->str);

  }
}
